/*
 * Copyright 2014 OGIS-RI All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.co.ogis_ri.optaplanner.vehiclerouting.domain;

import jp.co.ogis_ri.optaplanner.vehiclerouting.domain.core.ChainedTimeWindowedStep;
import jp.co.ogis_ri.optaplanner.vehiclerouting.domain.core.TimeWindowedStep;

public class Cargo extends ChainedTimeWindowedStep {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * 荷物の大きさ
	 */
	protected int size;

	/**
	 * 場所
	 */
	protected Location location;

	/**
	 * 荷物を積みこむのに必要な時間
	 */
	private int serviceDuration;

	/**
	 * 荷物の量を取得する
	 * 
	 * @return
	 */
	public int getSize() {
		return this.size;
	}

	/**
	 * 荷物の量を設定する
	 * 
	 * @param size
	 */
	public void setSize(final int size) {
		this.size = size;
	}

	/**
	 * このCargoの位置情報を取得する
	 */
	@Override
	public Location getLocation() {
		return this.location;
	}

	/**
	 * このCargoの位置情報を設定する
	 * 
	 * @param location
	 */
	public void setLocation(final Location location) {
		this.location = location;
	}

	/**
	 * 荷物を積みこむのに必要な時間を取得する
	 * 
	 * @return
	 */
	public int getServiceDuration() {
		return this.serviceDuration;
	}

	/**
	 * 荷物を積みこむのに必要な時間を設定する
	 * 
	 * @param serviceDuration
	 */
	public void setServiceDuration(final int serviceDuration) {
		this.serviceDuration = serviceDuration;
	}

	/**
	 * ひとつ前の地点からの距離を取得する
	 * 
	 * @return
	 */
	@Override
	public int getDistanceToPreviousStep() {
		if (this.previousStep == null) {
			return 0;
		}
		return this.getDistanceTo(this.previousStep);
	}

	/**
	 * ひとつ前の地点からの移動時間を取得する
	 * 
	 * @return
	 */
	@Override
	public int getTravelTimeToPreviousStep() {
		if (this.previousStep == null) {
			return 0;
		}
		return this.getTravelTimeTo(this.previousStep);
	}

	/**
	 * 指定された地点からの距離を取得する
	 * 
	 * @param step
	 * @return
	 */
	@Override
	public int getDistanceTo(final TimeWindowedStep step) {

		// 同じ地点の場合、移動距離は０になる
		if (this.location == step.getLocation()) {
			return 0;
		}

		return this.location.getDistance(step.getLocation());
	}

	/**
	 * 指定された地点からの移動時間を取得する
	 * 
	 * @param step
	 * @return
	 */
	@Override
	public int getTravelTimeTo(final TimeWindowedStep step) {

		// 同じ地点の場合、移動時間は０になる
		if (this.location == step.getLocation()) {
			return 0;
		}

		return this.location.getTravelTime(step.getLocation());
	}

	/**
	 * 
	 */
	@Override
	public String toString() {
		return this.location
				+ "(after "
				+ (this.previousStep == null ? "null" : this.previousStep
						.getLocation()) + ")";
	}

	/**
	 * この地点の出発時刻を取得する
	 */
	@Override
	public Integer getDepartureTime() {
		if (arrivalTime == null) {
			return null;
		}

		// 早く到着してしまうと荷物の準備完了時間まで待つしかないので
		// 到着時間と、荷物の準備完了時間、時間の大きい方に、
		// 荷物の積み込み時間を足す
		return Math.max(arrivalTime, readyTime) + serviceDuration;
	}
}
